unit FrmAbstraction;

{

   COPYRIGHT 2020 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://www.bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}


interface

uses

  Winapi.Windows, Winapi.Messages,

  System.SysUtils, System.Variants, System.Classes, System.StrUtils, System.Types, System.Generics.Collections,
  System.Generics.Defaults, System.math,

  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ComCtrls, KnobsWorms, Vcl.StdCtrls, Vcl.ExtCtrls, Vcl.Buttons,
  Vcl.Menus,

  KnobsUtils, knobs2013, KnobParser, Globals, AudioCaching,

  FrmStore, FrmFolderSelector;

type


  TFormAbstraction = class(TForm)
    PanelEditorTop: TPanel;
    SplitterModuleSelector: TSplitter;
    PanelEditorTopRight: TPanel;
    SpeedButtonChangeModuleSelector: TSpeedButton;
    PanelRunControls: TPanel;
    LabelPolyphony: TLabel;
    EditSearch: TEdit;
    UpDownPolyphony: TUpDown;
    ModuleSelector: TKnobsModuleSelector;
    PatchEditor: TKnobsWirePanel;
    abstraction_edit: TKnobsSelector;
    KnobsSelector1: TKnobsSelector;
    PopupMenuControl: TPopupMenu;
    MenuItemShowRunControls: TMenuItem;
    MenuItemShowMorphControls: TMenuItem;
    MenuItemShowMutationControls: TMenuItem;
    MenuItemShowTuningControls: TMenuItem;
    PopupMenuTextControl: TPopupMenu;
    TextMenuEdit: TMenuItem;
    TextMenuClear: TMenuItem;
    PopupMenuGridControl: TPopupMenu;
    MenuItemGridControlSaveToFile: TMenuItem;
    MenuItemGridControlLoadFromFile: TMenuItem;
    PopupMenuXYControl: TPopupMenu;
    MenuItemXYControlRandomize: TMenuItem;
    MenuItemXYControlExcludefromrandomization: TMenuItem;
    PopupMenuDataMaker: TPopupMenu;
    PopupMenuDataMakerSine: TMenuItem;
    PopupMenuDataMakerCosine: TMenuItem;
    PopupMenuDataMakerArcTan: TMenuItem;
    PopupMenuDataMakerBell: TMenuItem;
    PopupMenuDataMakerTri: TMenuItem;
    PopupMenuDataMakerSaw: TMenuItem;
    PopupMenuDataMakerSquare3: TMenuItem;
    PopupMenuDataMakerSquare10: TMenuItem;
    PopupMenuDataMakerSquare33: TMenuItem;
    PopupMenuDataMakerSquare50: TMenuItem;
    PopupMenuDataMakerSquare67: TMenuItem;
    PopupMenuDataMakerSquare90: TMenuItem;
    PopupMenuDataMakerSquare97: TMenuItem;
    PopupMenuDataMakerNoise: TMenuItem;
    N9: TMenuItem;
    PopupMenuDataMakerClear: TMenuItem;
    PopupMenuDataMakerSwapLeftRight: TMenuItem;
    PopupMenuDataMakerSwapTopBottom: TMenuItem;
    PopupMenuDataMakerRotate180: TMenuItem;
    PopupMenuDataMakerMakeloopable: TMenuItem;
    KMakefullscale1: TMenuItem;
    N16: TMenuItem;
    PopupMenuDataMakerSavetofile: TMenuItem;
    PopupMenuDataMakerLoadfromfile: TMenuItem;
    N22: TMenuItem;
    MenuItemDataMakerRandomize: TMenuItem;
    MenuItemDataMakerExcludefromrandomization: TMenuItem;
    PopupMenuEditor: TPopupMenu;
    MenuItemInsertmodule: TMenuItem;
    N17: TMenuItem;
    cut1: TMenuItem;
    copy1: TMenuItem;
    paste1: TMenuItem;
    pastevalues1: TMenuItem;
    delete1: TMenuItem;
    N18: TMenuItem;
    MenuItemEditorHistory: TMenuItem;
    selectall1: TMenuItem;
    invertselection1: TMenuItem;
    selectnone1: TMenuItem;
    PopupMenuSelector: TPopupMenu;
    dynamics1: TMenuItem;
    PopupMenuKnob: TPopupMenu;
    KnobMIDICC: TMenuItem;
    AssigntoCC: TMenuItem;
    UnassignCC: TMenuItem;
    N7: TMenuItem;
    SetKnobDefaultvalue: TMenuItem;
    LockKnob: TMenuItem;
    MenuItemKnobHistory: TMenuItem;
    N21: TMenuItem;
    MenuItemKnobRandomize: TMenuItem;
    MenuItemKnobExcludeFromRandomization: TMenuItem;
    PopupMenuModule: TPopupMenu;
    MenuItemModuleAlternates: TMenuItem;
    MenuItemModuleColor: TMenuItem;
    MenuItemModuleDefaultcolor: TMenuItem;
    N11: TMenuItem;
    MenuItemModuleHelp: TMenuItem;
    MenuItemModuleHistory: TMenuItem;
    N12: TMenuItem;
    MenuItemModuleCopy: TMenuItem;
    MenuItemModuleCut: TMenuItem;
    MenuItemModuleDelete: TMenuItem;
    MenuItemModulePasteParamsStrict: TMenuItem;
    MenuItemPasteParamsNonStrict: TMenuItem;
    N14: TMenuItem;
    MenuItemModuleSaveAsPreset: TMenuItem;
    MenuItemModuleLoadFromPreset: TMenuItem;
    N20: TMenuItem;
    MenuItemModuleRandomize: TMenuItem;
    MenuItemModuleExludeFromRandomization: TMenuItem;
    PopupMenuConnector: TPopupMenu;
    DisconnectConnector: TMenuItem;
    BreakWire: TMenuItem;
    DeleteConnector: TMenuItem;
    N5: TMenuItem;
    Redwires1: TMenuItem;
    Bluewires1: TMenuItem;
    Yellowwires1: TMenuItem;
    Makegreen1: TMenuItem;
    Customwirecolor1: TMenuItem;
    Defaultwirecolor1: TMenuItem;
    ColorDialogModule: TColorDialog;
    OpenDialogPreset: TOpenDialog;
    SaveDialogPreset: TSaveDialog;
    LedMidiRx: TKnobsLed;
    LedMidiActive: TKnobsLed;
    LabelTextMidiT: TLabel;
    LabelTextMidiR: TLabel;
    LabelTextMidiA: TLabel;
    LabelTextMidi: TLabel;
    LedMidiTx: TKnobsLed;
    LabelTextOsc: TLabel;
    LabelTextOscA: TLabel;
    LedOSCEnabled: TKnobsLed;
    LabelTextOScR: TLabel;
    LabelTextOscT: TLabel;
    LedOSCRx1: TKnobsLed;
    LedOSCTx1: TKnobsLed;
    procedure ModuleSelectorModuleButtonClick(const aSender: TObject; aModuleclass: Integer);
    procedure FormCreate(Sender: TObject);
    procedure DisconnectConnectorClick(Sender: TObject);
    procedure BreakWireClick(Sender: TObject);
    procedure DeleteConnectorClick(Sender: TObject);
    procedure Redwires1Click(Sender: TObject);
    procedure Bluewires1Click(Sender: TObject);
    procedure Yellowwires1Click(Sender: TObject);
    procedure Makegreen1Click(Sender: TObject);
    procedure Customwirecolor1Click(Sender: TObject);
    procedure Defaultwirecolor1Click(Sender: TObject);
    procedure MenuItemModuleColorClick(Sender: TObject);
    procedure MenuItemModuleDefaultcolorClick(Sender: TObject);
    procedure MenuItemModuleHelpClick(Sender: TObject);
    procedure MenuItemModuleCopyClick(Sender: TObject);
    procedure MenuItemModuleCutClick(Sender: TObject);
    procedure MenuItemModuleDeleteClick(Sender: TObject);
    procedure MenuItemModulePasteParamsStrictClick(Sender: TObject);
    procedure MenuItemPasteParamsNonStrictClick(Sender: TObject);
    procedure MenuItemModuleSaveAsPresetClick(Sender: TObject);
    procedure MenuItemModuleLoadFromPresetClick(Sender: TObject);
    procedure MenuItemModuleRandomizeClick(Sender: TObject);
    procedure MenuItemModuleExludeFromRandomizationClick(Sender: TObject);
    procedure abstraction_editClick(Sender: TObject);
    procedure KnobsSelector1Click(Sender: TObject);
    procedure UpDownPolyphonyClick(Sender: TObject; Button: TUDBtnType);
    procedure FormDestroy(Sender: TObject);
  private
    FSynthModule          : TObject;
    FPatchModule          : TKnobsCustomModule;
    FFormStore            : TFormStore;
    FFormWren             : TForm;
    FStandardColors       : Boolean;
    FMidiChannel          : Integer;
    FPatchReader          : IKnobsPatchReader;
    FPatchWriter          : IKnobsPatchWriter;
    FPopupSender          : TControl;
    FLastCC               : Byte;
    FLastFrozenCC         : Byte;
    FWarnOnStructMismatch : Boolean;
    FModuleColor          : TColor;
  private
    procedure   FixModulePolyphony;
    function    GetPolyphony  : Integer;
    procedure   SetPolyphony  ( aValue: Integer);
    function    GetVersion    : Integer;
    function    GetModuleCount: Integer;
    function    GetModule     ( anIndex: Integer): TKnobsCustomModule;
    function    GetWireCount  : Integer;
    function    GetWire       ( anIndex: Integer): TKnobsWire;
  private
    procedure   Log   ( aLogClass: Integer; const aMsg: string);
    procedure   LogFmt( aLogClass: Integer; const aFmt: string; const anArgs: array of const);
    procedure   LoadPatchPreset;
    procedure   SavePatchPreset;
    procedure   LoadPreset( aModule: TKnobsCustomModule);
    procedure   SavePreset( aModule: TKnobsCustomModule);
    procedure   DeleteModules;
    procedure   CutModules;
    procedure   CopyModules;
 // procedure   PasteModules;
    procedure   PasteParameters( StrictCopy: Boolean);
    function    GetDefaultModuleColor    ( const aSender: TObject; const aModuleType: TKnobsModuleType): TColor;
    procedure   ShowModuleHelp( aModule: TKnobsCustomModule);
    procedure   ConnectorDelete          ( const aConnector    : TKnobsConnector    );
    procedure   ConnectorBreak           ( const aConnector    : TKnobsConnector    );
    procedure   ConnectorDisconnect      ( const aConnector    : TKnobsConnector    );
    procedure   SetWireColor             ( const aConnector    : TKnobsConnector; const aColor: TColor);
    procedure   SetWireColorDefault      ( const aConnector    : TKnobsConnector    );
    procedure   SetWireColorCustom       ( const aConnector    : TKnobsConnector    );
    procedure   ModuleCut                ( const aModule       : TKnobsModule       );
    procedure   ModuleCopy               ( const aModule       : TKnobsModule       );
    procedure   ModuleDelete             ( const aModule       : TKnobsModule       );
    procedure   ModulePasteParams        ( const aModule       : TKnobsModule; StrictCopy: Boolean);
    procedure   ChangeModuleColor        ( const aModule       : TKnobsModule       );
    procedure   ModuleColorDefault       ( const aModule       : TKnobsModule       );
    procedure   ModuleRandomize          ( const aModule       : TKnobsModule       );
    procedure   ModuleRandomizationToggle( const aModule       : TKnobsModule       );
    procedure   DoSelectorMenuItemClick( aSender: TObject);
    procedure   AddSelectorMenuItem( anIndex, aShortcut: Integer; const aName: string);
    procedure   RemoveSelectorMenuItems;
    procedure   DoAlternateModuleMenuItemClick( aSender: TObject);
    procedure   AddModuleAlternateMenuItem( anIndex: Integer; const aCurrent, aName: string; aModuleType: Integer);
    procedure   RemoveModuleAlternateMenuItems;
    procedure   DoHistoryChanged        ( const aSender: TObject; anUndoCount, aRedoCount, aSavedMarker: Integer);
    procedure   DoValueChanged          ( const aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal, IsAutomation: Boolean);
    procedure   DoTextChanged           ( const aSender: TObject; const aPath, aValue: string);
    procedure   DoRecompile             ( const aSender: TKnobsWirePanel; ModulesChanged: Boolean);
    procedure   DoHistoryItemClicked    ( aSender: TObject);
    procedure   DoShowPopupMenu         ( const aSender: TKnobsWirePanel; const aControl: TControl);
    procedure   DoUnFocus               ( const aSender: TObject);
    function    DoLoadDataGraph         ( const aSender: TObject): string;
    procedure   DoSaveDataGraph         ( const aSender: TObject; const aValue: string);
    function    DoLoadGridControl       ( const aSender: TObject): string;
    procedure   DoSaveGridControl       ( const aSender: TObject; const aValue: string);
    procedure   DoActiveVariationChanged( const aSender: TObject; aVariation: Integer);
    procedure   DoLoadCaptions          ( const aSender: TObject; const aModule: TKnobsCustomModule; const aControl: TKnobsSelector; const aDependency: TKnobsValuedControl);
    procedure   DoEditHistoryChanged    ( aSender: TObject);
    procedure   DoModulesRenamed        ( const aSender: TObject; const anOldNames, aNewNames: TStringArray);
    procedure   DoModulesRenamedFriendly( const aSender: TObject; const anOldNames, aNewNames: TStringArray);
    procedure   DoAudioChanged          ( const aSender: TObject; const anOldFileName, aNewFileName: string);
    procedure   DoBeginStateChange      ( const aSender: TObject);
    procedure   DoEndStateChange        ( const aSender: TObject);
  private
    procedure   DoOnValuedControlRemoved( const aSender: TKnobsValuedControl);
    procedure   SelectNone;
    procedure   AddModule( aModuleType: TKnobsModuleType);
    procedure   ChangeModuleType( var aTarget: TKnobsCustomModule; aWantedType: TKnobsModuleType);
  public
    procedure   MakeSpedUpSlow;
    procedure   FixAllWires;
    procedure   SortGraph;
    procedure   BuildModuleIndex;
    procedure   FixAllSynthParams( IsAutomation: Boolean);
    function    FindModule( const aName: string): TKnobsCustomModule;
    function    AbstractionToStr: string;
    procedure   LoadPatchFromString( const aPatch: string);
    procedure   BuildIoMaps( var aMaps: TKnobsIoArray);
  public
    property    FormWren                  : TForm              read FFormWren             write FFormWren;
    property    SynthModule               : TObject            read FSynthModule          write FSynthModule;
    property    PatchModule               : TKnobsCustomModule read FPatchModule          write FPatchModule;
    property    FormStore                 : TFormStore         read FFormStore            write FFormStore;
    property    StandardColors            : Boolean            read FStandardColors       write FStandardColors;
    property    MidiChannel               : Integer            read FMidiChannel          write FMidiChannel;
    property    Polyphony                 : Integer            read GetPolyphony          write SetPolyphony;
    property    WarnOnStructMismatch      : Boolean            read FWarnOnStructMismatch write FWarnOnStructMismatch;
    property    ModuleColor               : TColor             read FModuleColor          write FModuleColor;
    property    Editor                    : TKnobsWirePanel    read PatchEditor;
    property    Version                   : Integer            read GetVersion;
    property    ModuleCount               : Integer            read GetModuleCount;
    property    Module[ anIndex: Integer] : TKnobsCustomModule read GetModule;
    property    WireCount                 : Integer            read GetWireCount;
    property    Wire  [ anIndex: Integer] : TKnobsWire         read GetWire;
  end;



implementation



uses

  FrmMain,

  module_defs, module_implementations;


// User area


  // //////// //////// //////// //////// //////// //////// //////// //////// //////// //////// //////// //
  //                                                                                                    //
  // The following two are here to avoid recursive unit usage, FrmMain and module_xxx can not be in the //
  // interface uses section, so for the synth module and the main form we need generic types that then  //
  // are typecasted to the real thing by the functions below.                                           //
  //                                                                                                    //
                                                                                                        //
    function  AsSynthModule( aSynthModule: TObject): TModAbstraction;                                   //
    begin                                                                                               //
      Result := aSynthModule as TModAbstraction;                                                        //
    end;                                                                                                //
                                                                                                        //
    function  AsFormWren( aForm: TForm): TFormWren;                                                     //
    begin                                                                                               //
      Result := aForm as TFormWren;                                                                     //
    end;                                                                                                //
                                                                                                        //
  //                                                                                                    //
  // //////// //////// //////// //////// //////// //////// //////// //////// //////// //////// //////// //


//  private

    procedure   TFormAbstraction.FixModulePolyphony;
    var
      i : Integer;
    begin
      for i := 0 to FPatchModule.ControlCount - 1
      do begin
        if   FPatchModule.Controls[ i] is TKnobsValuedControl
        and  SameText( TKnobsValuedControl( FPatchModule.Controls[ i]).Name, 'abstraction_voicecount')
        then begin
          TKnobsValuedControl( FPatchModule.Controls[ i]).KnobPosition := Polyphony - 1;
          Break;
        end;
      end;
    end;


    function    TFormAbstraction.GetPolyphony: Integer;
    begin
      Result := PatchEditor.Polyphony;
    end;


    procedure   TFormAbstraction.SetPolyphony( aValue: Integer);
    begin
      aValue := Clip( aValue, MIN_POLYPHONY, MAX_POLYPHONY);

      if   aValue <> Polyphony
      then begin

        PatchEditor.Polyphony := aValue;
        UpDownPolyphony.Position := aValue;
        FixModulePolyphony;
      end;

      LabelPolyphony.Caption := Format( 'poly:%d', [ Polyphony], AppLocale);
    end;


    function    TFormAbstraction.GetVersion: Integer;
    begin
      Result := PatchEditor.Version;
    end;


    function    TFormAbstraction.GetModuleCount: Integer;
    begin
      Result := PatchEditor.ModuleCount;
    end;


    function    TFormAbstraction.GetModule( anIndex: Integer): TKnobsCustomModule;
    begin
      Result := PatchEditor.Module[ anIndex];
    end;


    function    TFormAbstraction.GetWireCount: Integer;
    begin
      Result := PatchEditor.WireCount;
    end;


    function    TFormAbstraction.GetWire( anIndex: Integer): TKnobsWire;
    begin
      Result := PatchEditor.Wire[ anIndex];
    end;


//  private

    procedure   TFormAbstraction.DoOnValuedControlRemoved( const aSender: TKnobsValuedControl);
    begin
      // todo: MIDI CC mapping
//      UnAssignMidiCC( aSender);
//
//      if   Assigned( FormLastEdits)
//      then FormLastEdits.RemoveValuedControl( aSender);
    end;


    procedure   TFormAbstraction.Log( aLogClass: Integer; const aMsg: string);
    var
      aForm : TFormWren;
    begin
      aForm := AsFormWren( FFormWren);

      if   Assigned( aForm)
      then aForm.LogFmt( aLogClass, 'TFormAbstraction : %s', [ aMsg]);
    end;


    procedure   TFormAbstraction.LogFmt( aLogClass: Integer; const aFmt: string; const anArgs: array of const);
    begin
      Log( aLogClass, Format( aFmt, anArgs, AppLocale));
    end;


    procedure   TFormAbstraction.LoadPatchPreset;
    begin
      try
        PatchEditor.SelectAllModules;

        with OpenDialogPreset
        do begin
          FileName   := ChangeFileExt( ExtractFileName( PatchEditor.Filename), EXT_PRESET);
          InitialDir := FormFolderSelector.FolderPresets;

          if   Execute
          then begin
            FPatchReader.WarnOnStructMismatch := Self.WarnOnStructMismatch;
            FPatchReader.ParamsFromFile( FileName, PatchEditor, DO_STRICT);
          end;
        end;
      finally
        PatchEditor.UnSelectAllModules;
      end;
    end;


    procedure   TFormAbstraction.SavePatchPreset;
    begin
      if   not Assigned( FormFolderSelector)
      then Exit;

      try
        PatchEditor.SelectAllModules;

        with SaveDialogPreset
        do begin
          FileName   := ChangeFileExt( ExtractFileName( PatchEditor.Filename), EXT_PRESET);
          InitialDir := FormFolderSelector.FolderPresets;

          if   Execute
          then begin
            try
              FPatchWriter.WriteFile( PatchEditor, FileName, wmParams);
            except
              on E: Exception
              do begin
                LogFmt( LC_GENERAL, 'TFormAbstraction - Failed to write "%s", reason: %s', [ FileName, E.ToString]);
              end;
            end;
          end;
        end;
      finally
        PatchEditor.UnSelectAllModules;
      end;
    end;


    procedure   TFormAbstraction.LoadPreset( aModule: TKnobsCustomModule);
    begin
      if   not Assigned( FormFolderSelector)
      then Exit;

      if   Assigned( aModule)
      then begin
        with OpenDialogPreset
        do begin
          if   aModule.TemplateName <> ''
          then begin
            FileName   := ChangeFileExt( ExtractFileName( aModule.TemplateName), EXT_PRESET);
            InitialDir := ExtractFilePath( aModule.TemplateName);
          end
          else begin
            FileName   := ChangeFileExt( aModule.Title, EXT_PRESET);
            InitialDir := FormFolderSelector.FolderPresets;
          end;

          if   Execute
          then begin
            aModule.FixSelection;
            aModule.TemplateName := FileName;
            FPatchReader.WarnOnStructMismatch := Self.WarnOnStructMismatch;
            FPatchReader.ParamsFromFile( FileName, PatchEditor, DO_STRICT);
          end;
        end;
      end;
    end;


    procedure   TFormAbstraction.SavePreset( aModule: TKnobsCustomModule);
    begin
      if   not Assigned( FormFolderSelector)
      then Exit;

      if   Assigned( aModule)
      then begin
        with SaveDialogPreset
        do begin
          if   aModule.TemplateName <> ''
          then begin
            FileName   := ChangeFileExt( ExtractFileName( aModule.TemplateName), EXT_PRESET);
            InitialDir := ExtractFilePath( aModule.TemplateName);
          end
          else begin
            FileName   := ChangeFileExt( aModule.Title, EXT_PRESET);
            InitialDir := FormFolderSelector.FolderPresets;
          end;

          if   Execute
          then begin
            aModule.FixSelection;
            aModule.TemplateName := FileName;

            try
              FPatchWriter.WriteFile( PatchEditor, FileName, wmParams);
            except
              on E: Exception
              do LogFmt( LC_GENERAL, 'Failed to write "%s", reason: %s', [ FileName, E.ToString]);
            end;
          end;
        end;
      end;
    end;


    procedure   TFormAbstraction.DeleteModules;
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.FreeSelectedModules( DoOnValuedControlRemoved);
    end;


    procedure   TFormAbstraction.CutModules;
    begin
      CopyModules;
      DeleteModules;
    end;


    procedure   TFormAbstraction.CopyModules;
    begin
      if   Assigned( FPatchWriter)
      then FPatchWriter.WriteToClipboard( PatchEditor, wmSelected);
    end;


//    procedure   TFormAbstraction.PasteModules;
//    begin
//      if   Assigned( FPatchReader)
//      then FPatchReader.ReadFromClipboard( PatchEditor, DO_STRICT, DO_DRAG, NO_COPY_MIDI, NO_COPY_TUNING, NO_COPY_RATE, rmAppend);
//    end;


    procedure   TFormAbstraction.PasteParameters( StrictCopy: Boolean);
    begin
      if   Assigned( FPatchReader)
      then begin
        FPatchReader.WarnOnStructMismatch := Self.WarnOnStructMismatch;
        FPatchReader.ParamsFromClipboard( PatchEditor, StrictCopy);
      end;
    end;


    function    TFormAbstraction.GetDefaultModuleColor( const aSender: TObject; const aModuleType: TKnobsModuleType): TColor;
    begin
      if   Assigned( FormStore)
      then Result := FormStore.DefaultModuleColor( aModuleType, ModuleColor)
      else Result := ModuleColor;
    end;


    procedure   TFormAbstraction.ShowModuleHelp( aModule: TKnobsCustomModule);
    var
      S : PChar;
    begin                                                       // GetText uses StrNew ...
      S := aModule.Docs.GetText;                                // ... so we need use StrDispose to free it -
      ShowMessage( FixHtmlTokens( aModule.Title + ^M^M + S));   // this is what the help on GetText sez too, right.
      StrDispose( S);                                           // ok.
    end;


    procedure   TFormAbstraction.ConnectorDelete( const aConnector: TKnobsConnector);
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.ConnectorDelete( aConnector);
    end;


    procedure   TFormAbstraction.ConnectorBreak( const aConnector: TKnobsConnector);
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.ConnectorBreak( aConnector);
    end;


    procedure   TFormAbstraction.ConnectorDisconnect( const aConnector: TKnobsConnector);
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.ConnectorDisconnect( aConnector);
    end;


    procedure   TFormAbstraction.SetWireColor( const aConnector: TKnobsConnector; const aColor: TColor);
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.SetCustomWireColor( aConnector, aColor);
    end;


    procedure   TFormAbstraction.SetWireColorDefault( const aConnector: TKnobsConnector);
    begin
      if   Assigned( PatchEditor)
      then PatchEditor.SetWireColorDefault( aConnector);
    end;


    procedure   TFormAbstraction.SetWireColorCustom( const aConnector: TKnobsConnector);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aConnector)
      then begin
        with ColorDialogModule
        do begin
          Color := aConnector.WireColor;

          if   Execute
          then PatchEditor.SetCustomWireColor( aConnector, Color);
        end;
      end;
    end;


    procedure   TFormAbstraction.ModuleCut( const aModule: TKnobsModule);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aModule)
      then begin
        aModule.FixSelection;
        CutModules;
      end;
    end;


    procedure   TFormAbstraction.ModuleCopy( const aModule: TKnobsModule);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aModule)
      then begin
        aModule.FixSelection;
        CopyModules;
      end;
    end;


    procedure   TFormAbstraction.ModuleDelete( const aModule: TKnobsModule);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aModule)
      then begin
        aModule.FixSelection;
        DeleteModules;
      end;
    end;


    procedure   TFormAbstraction.ModulePasteParams( const aModule: TKnobsModule; StrictCopy: Boolean);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aModule)
      then PasteParameters( StrictCopy);
    end;


    procedure   TFormAbstraction.ChangeModuleColor( const aModule: TKnobsModule);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( aModule)
      then begin
        aModule.FixSelection;

        with ColorDialogModule
        do begin
          Color := aModule.Color;

          if   Execute
          then PatchEditor.SetSelectedModuleColor( Color);
        end;
      end;
    end;


    procedure   TFormAbstraction.ModuleColorDefault( const aModule: TKnobsModule);
    begin
      if   Assigned( PatchEditor)
      and  Assigned( FormStore  )
      and  Assigned( aModule    )
      then begin
        aModule.FixSelection;
        PatchEditor.SetSelectedModuleColorDefault( GetDefaultModuleColor);
      end;
    end;


    procedure   TFormAbstraction.ModuleRandomize( const aModule: TKnobsModule);
    begin
      if   Assigned( aModule)
      then begin
        if   not aModule.Selected
        then aModule.SelectUnique;

        // SetRandomValue;
      end;
    end;


    procedure   TFormAbstraction.ModuleRandomizationToggle( const aModule: TKnobsModule);
    begin
      if   Assigned( aModule)
      then begin
        if   not aModule.Selected
        then aModule.SelectUnique;

        // SetAllowRandomization( not aModule.AllowRandomization)
      end;
    end;


    procedure   TFormAbstraction.DoSelectorMenuItemClick( aSender: TObject);
    var
      aSelector : TKnobsSelector;
      aMenuItem : TMenuItem;
    begin
      aMenuItem := aSender as TMenuItem;

      if   Assigned( aMenuItem)
      then begin
        aSelector := TKnobsSelector( FPopupSender);
        aSelector.BeginStateChange;

        try
          aSelector.KnobPosition := aMenuItem.Tag;
        finally
          aSelector.EndStateChange;
        end;

        aSelector.FixPeers;
      end;
    end;


    procedure   TFormAbstraction.AddSelectorMenuItem( anIndex, aShortcut: Integer; const aName: string);
    var
      aMenuItem : TMenuItem;
    begin
      aMenuItem := TMenuItem.Create( PopupMenuSelector);

      with aMenuItem
      do begin

        if   aName <> '-'
        then begin
          Caption    := Format( '%s (%d) %s', [ IndexToMenuShortCut( aShortcut), anIndex, aName], AppLocale);
          OnClick    := DoSelectorMenuItemClick;
          Tag        := anIndex;
          ImageIndex := anIndex;
        end
        else Caption := aName;
      end;

      PopupMenuSelector.Items.Add( aMenuItem);
    end;


    procedure   TFormAbstraction.RemoveSelectorMenuItems;
    var
      i : Integer;
    begin
      with PopupMenuSelector
      do begin
        for i := Items.Count - 1 downto 0
        do  Items.Delete( i);
      end;
    end;


    procedure   TFormAbstraction.DoAlternateModuleMenuItemClick( aSender: TObject);
    var
      aMenuItem   : TMenuItem;
      aModule     : TKnobsCustomModule;
      CurrentType : Integer;
      WantedType  : Integer;
    begin
      if   Assigned( PatchEditor)
      then begin
        aMenuItem := aSender as TMenuItem;

        if   Assigned( aMenuItem)
        then begin
          aModule := TKnobsCustomModule( FPopupSender);

          if   Assigned( aModule)
          then begin
            CurrentType := aModule.ModuleType;
            WantedType  := aMenuItem.Tag;

            if   CurrentType <> WantedType
            then ChangeModuleType( aModule, WantedType);  // This Frees aModule, do not use it after that ...
          end;
        end;
      end;
    end;


    procedure   TFormAbstraction.AddModuleAlternateMenuItem( anIndex: Integer; const aCurrent, aName: string; aModuleType: Integer);
    var
      aMenuItem : TMenuItem;
    begin
      aMenuItem := TMenuItem.Create( MenuItemModuleAlternates);

      with aMenuItem
      do begin
        if   aName <> '-'
        then begin
          Caption := Format( '%s%s', [ IndexToMenuShortCut( anIndex), aName], AppLocale);
          Tag     := aModuleType;
          OnClick := DoAlternateModuleMenuItemClick;
        end
        else Caption := aName;
      end;

      MenuItemModuleAlternates.Add( aMenuItem);
      aMenuItem.Checked := SameText( aCurrent, aName);
    end;


    procedure   TFormAbstraction.RemoveModuleAlternateMenuItems;
    var
      i : Integer;
    begin
      with MenuItemModuleAlternates
      do begin
        for i := Count - 1 downto 0
        do  Delete( i);
      end;
    end;


    procedure   TFormAbstraction.DoHistoryChanged( const aSender: TObject; anUndoCount, aRedoCount, aSavedMarker: Integer);
    begin
//      MenuEditUndo.Enabled := anUndoCount  > 0;
//      MenuEditRedo.Enabled := aRedoCount   > 0;
//      PatchChanged         := anUndoCount <> aSavedMarker;
    end;


    procedure   TFormAbstraction.DoValueChanged( const aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal, IsAutomation: Boolean);
    // An editor module's control changed value, to be reflected in a synth setting
    // Must be passed on to the Abstraction module
    begin
      AsSynthModule( FSynthModule).ValueChanged( aSender, aPath, aControlType, aValue, IsFinal, IsAutomation);
    end;


    procedure   TFormAbstraction.DoTextChanged( const aSender: TObject; const aPath, aValue: string);
    // Must be passed on to the Abstraction module
    begin
      AsSynthModule( FSynthModule).TextChanged( aSender, aPath, aValue);
    end;


    procedure   TFormAbstraction.DoRecompile( const aSender: TKnobsWirePanel; ModulesChanged: Boolean);
    // Must be passed on to the Abstraction module
    begin
      Log( LC_DEBUG, 'TFormAbstraction.DoRecompile');
      AsSynthModule( FSynthModule).CompilePatch( ModulesChanged);
    end;


    procedure   TFormAbstraction.DoHistoryItemClicked( aSender: TObject);
    begin
      if   Assigned( FPopupSender)
      and  IsUndoable( FPopupSender)
      and  ( aSender is TMenuItem)
      then ( FPopupSender as IKnobsUndoable).UndoItem( TMenuItem( aSender).Tag);
    end;


    procedure   TFormAbstraction.DoShowPopupMenu( const aSender: TKnobsWirePanel; const aControl: TControl);

      procedure PopUp( aPopUp: TPopupMenu; aControl: TControl);
      var
        aPoint : TPoint;
      begin
        FPopupSender := aControl;
        aPoint := Mouse.CursorPos;
        aPopup.Popup( aPoint.X + 10, aPoint.Y + 10);
      end;

      procedure AddHistoryItems( aMenuItem: TMenuItem; anUndoable: IKnobsUndoable; aMaxItems: Integer);
      var
        aList        : TStringList;
        i            : Integer;
        aSubMenuItem : TMenuItem;
      begin
        if   Assigned( anUndoable)
        then begin
          with aMenuItem
          do begin
            for i := Count - 1 downto 0
            do Delete( i);
          end;

          aMenuItem.Enabled := False;
          aList             := anUndoable.CreateStringList( aMaxItems); // max 36 items - 0 .. 9. A .. Z

          try
            for i := 0 to aList.Count - 1
            do begin
              aSubMenuItem := TMenuItem.Create( aMenuItem);

              with aSubMenuItem
              do begin
                Caption := Format( '%s%s', [ IndexToMenuShortCut( i), aList[ i]], AppLocale);
                Tag     := i;
                OnClick := DoHistoryItemClicked;
                Enabled := True;
              end;

              aMenuItem.Add( aSubMenuItem);
            end;
          finally
            aList.DisposeOf;
          end;

          aMenuItem.Enabled := aMenuItem.Count > 0;
        end;
      end;


      procedure ShowPopupModule( aPopup: TPopupMenu; aControl: TControl);
      var
        aModule    : TKnobsCustomModule;
        Alternates : string;
        aCurrent   : string;
        aType      : Integer;
        aName      : string;
        aList      : TStringList;
        i          : Integer;
        p          : Integer;
      begin
        RemoveModuleAlternateMenuItems;
        aModule    := TKnobsCustomModule( aControl);
        Alternates := aModule.ModuleAlternateTypes;
        aCurrent   := LookupModuleTypeName( aModule.ModuleType);
        aList      := Explode( Alternates, ' ');
        p          := 0;

        try
          try
            for i := 0 to aList.Count - 1
            do begin
              aType := StrToIntDef( aList[ i], -1);

              if   atype >= 0
              then begin
                aName  := LookupModuleTypeName( aType);

                if   aName <> ''
                then begin
                  AddModuleAlternateMenuItem( p, aCurrent, aName, aType);
                  Inc( p);
                end;
              end;
            end;
          finally
            aList.DisposeOf;
          end;
        except
          on E: Exception
          do KilledException( E);
        end;

        MenuItemModuleAlternates             .Enabled := p > 0;
        MenuItemModuleExludeFromRandomization.Checked := not aModule.AllowRandomization;

        if   IsUndoable( aControl)
        then AddHistoryItems( MenuItemModuleHistory, aControl as IKnobsUndoable, MODULE_HISTORY_COUNT);

        PopUp( aPopup, aControl);
      end;


      procedure ShowPopupSelector( aPopup: TPopupMenu; aControl: TControl);
      var
        i         : Integer;
        p         : Integer;
        aMenuItem : TMenuItem;
        anAuto    : IKnobsAutomatable;
        aRand     : IKnobsVariation;
        aSelect   : TKnobsSelector;
      begin
        if   Assigned( aControl)
        and  ( aControl is TKnobsSelector)
        then begin
          aSelect       := TKnobsSelector( aControl);
          p             := 0;
          aPopup.Images := aSelect.Glyphs;
          RemoveSelectorMenuItems;

          if   IsAutomatable( aControl)
          and  Assigned( aSender)
          then begin
            anAuto    := aControl as IKnobsAutomatable;
            aMenuItem := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              if   anAuto.AssignedMIDICC = 0
              then Caption := Format( '%sMIDI CC ...'     , [ IndexToMenuShortCut( p)                       ], AppLocale)
              else Caption := Format( '%sMIDI CC [%d] ...', [ IndexToMenuShortCut( p), anAuto.AssignedMIDICC], AppLocale);

           // todo: OnClick := KnobMIDICCClick;
              Enabled := anAuto.AllowAutomation;
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            Inc( p);
            FLastFrozenCC := FlastCC;
            aMenuItem     := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              Caption := Format( '%sAssign CC %d', [ IndexToMenuShortCut( p), FlastFrozenCC], AppLocale);
           // todo: OnClick := AssigntoCCClick;
              Enabled := ( FlastCC > 0) and anAuto.AllowAutomation;
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            Inc( p);
            aMenuItem := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              Caption := Format( '%sUnassign CC %d', [ IndexToMenuShortCut( p), anAuto.AssignedMIDICC], AppLocale);
           // todo: OnClick := UnassignCCClick;
              Enabled := anAuto.AllowAutomation and ( anAuto.AssignedMIDICC > 0);
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            Inc( p);

            if   p > 0
            then AddSelectorMenuItem( 0, -1, '-');
          end;

          if   IsVariation( aControl)
          and  Assigned( aSender)
          then begin
            aRand     := aControl as IKnobsVariation;
            aMenuItem := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              Caption := Format( '%sRandomize', [ IndexToMenuShortCut( p)], AppLocale);
           // todo: OnClick := MenuItemKnobRandomizeClick;
              Enabled := aRand.CanAllowRandomization and aRand.AllowRandomization;
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            Inc( p);

            aMenuItem := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              Caption := Format( '%sExclude from randomization', [ IndexToMenuShortCut( p)], AppLocale);
              Checked := not aRand.AllowRandomization;
           // todo: OnClick := ToggleSelectorRandomizationClick;
              Enabled := aRand.CanAllowRandomization;
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            Inc( p);

            AddSelectorMenuItem( 0, -1, '-');
          end;

          if   IsUndoable( aControl)
          and  Assigned( aSender)
          and  ( aSelect.StepCount > 1)
          then begin
            aMenuItem := TMenuItem.Create( PopupMenuSelector);

            with aMenuItem
            do begin
              Caption := Format( '%sHistory', [ IndexToMenuShortCut( p)], AppLocale);
              Inc( p);
              OnClick := nil;
            end;

            PopupMenuSelector.Items.Add( aMenuItem);
            AddHistoryItems( aMenuItem, aControl as IKnobsUndoable, CONTROL_HISTORY_COUNT);
            AddSelectorMenuItem( 0, -1, '-');
          end;

          for i := 0 to aSelect.Captions.Count - 1
          do begin
            if   aSelect.Captions[ i] = '-'
            then AddSelectorMenuItem( -1, -1, '-')
            else begin
              AddSelectorMenuItem( i, p, aSelect.Captions[ i]);
              Inc( p);
            end;
          end;

          PopUp( aPopup, aControl);
        end;
      end;


      procedure ShowPopupKnob( aPopup: TPopupMenu; aControl: TControl);
      begin
        if   aControl is TKnobsKnob
        then LockKnob.Checked := TKnobsKnob( aControl).Locked
        else LockKnob.Visible := False;

        if   IsAutomatable( aControl)
        then begin
          if  ( aControl as IKnobsAutomatable).AssignedMIDICC = 0
          then KnobMIDICC.Caption := Format( '%sMIDI CC ...'     , [ IndexToMenuShortcut( 0)], AppLocale)
          else KnobMIDICC.Caption := Format( '%sMIDI CC [%d] ...', [ IndexToMenuShortcut( 0), ( aControl as IKnobsAutomatable).AssignedMIDICC], AppLocale);

          KnobMIDICC.Enabled := ( aControl as IKnobsAutomatable).AllowAutomation;
          FLastFrozenCC      := FLastCC;
          AssigntoCC.Caption := Format( '%sAssign to CC %d', [ IndexToMenuShortcut( 1), FlastFrozenCC], AppLocale);
          AssigntoCC.Enabled := ( FLastFrozenCC > 0) and ( aControl as IKnobsAutomatable).AllowAutomation;
          UnAssignCC.Caption := Format( '%sUnassign CC %d', [ IndexToMenuShortcut( 2), ( aControl as IKnobsAutomatable).AssignedMIDICC], AppLocale);
          UnAssignCC.Enabled := ( aControl as IKnobsAutomatable).AssignedMIDICC > 0;
        end;

        if   IsVariation( aControl)
        then begin
          MenuItemKnobExcludeFromRandomization.Checked := not ( aControl as IKnobsVariation).AllowRandomization;
          MenuItemKnobExcludeFromRandomization.Enabled := ( aControl as IKnobsVariation).CanAllowRandomization;
          MenuItemKnobRandomize               .Enabled := ( aControl as IKnobsVariation).CanAllowRandomization and ( aControl as IKnobsVariation).AllowRandomization;
        end
        else begin
          MenuItemKnobExcludeFromRandomization.Checked := True;
          MenuItemKnobExcludeFromRandomization.Enabled := False;
          MenuItemKnobRandomize               .Enabled := False;
        end;

        if   IsUndoable( aControl)
        then AddHistoryItems( MenuItemKnobHistory, aControl as IKnobsUndoable, CONTROL_HISTORY_COUNT);

        PopUp( aPopup, aControl);
      end;


      procedure ShowPopupXYControl( aPopup: TPopupMenu; aControl: TControl);
      begin
        if   IsVariation( aControl)
        then begin
          MenuItemXYControlExcludeFromRandomization.Checked := not ( aControl as IKnobsVariation).AllowRandomization;
          MenuItemXYControlExcludeFromRandomization.Enabled := ( aControl as IKnobsVariation).CanAllowRandomization;
          MenuItemXYControlRandomize               .Enabled := ( aControl as IKnobsVariation).CanAllowRandomization and ( aControl as IKnobsVariation).AllowRandomization;
        end
        else begin
          MenuItemXYControlExcludeFromRandomization.Checked := True;
          MenuItemXYControlExcludeFromRandomization.Enabled := False;
          MenuItemXYControlRandomize               .Enabled := False;
        end;

        PopUp( aPopup, aControl);
      end;


      procedure ShowPopupDataMaker( aPopup: TPopupMenu; aControl: TControl);
      begin
        if   IsVariation( aControl)
        then begin
          MenuItemDataMakerExcludeFromRandomization.Checked := not ( aControl as IKnobsVariation).AllowRandomization;
          MenuItemDataMakerExcludeFromRandomization.Enabled := ( aControl as IKnobsVariation).CanAllowRandomization;
          MenuItemDataMakerRandomize               .Enabled := ( aControl as IKnobsVariation).CanAllowRandomization and ( aControl as IKnobsVAriation).AllowRandomization;
        end
        else begin
          MenuItemDataMakerExcludeFromRandomization.Checked := True;
          MenuItemDataMakerExcludeFromRandomization.Enabled := False;
          MenuItemDataMakerRandomize               .Enabled := False;
        end;

        PopUp( aPopup, aControl);
      end;


      procedure ShowPopupEditor( aPopup: TPopupMenu; aControl: TControl);
      begin
        if   IsUndoable( aControl)
        then AddHistoryItems( MenuItemEditorHistory, aControl as IKnobsUndoable, EDITOR_HISTORY_COUNT);

        PopUp( aPopup, aControl);
      end;


      procedure ShowPopupTextControl( aPopup: TPopupMenu; aControl: TControl);
      begin
        TextMenuClear.Enabled := TKnobsTextControl( aControl).AllowClear;
        PopUp( aPopup, aControl);
      end;

      procedure ShowPopupSimple( aPopup: TPopupMenu; aControl: TControl);
      begin
        PopUp( aPopup, aControl);
      end;

    begin
      if      aControl is TKnobsConnector
      then ShowPopupSimple( PopupMenuConnector, aControl)
      else if aControl is TKnobsDataMaker
      then ShowPopupDataMaker( PopupMenuDataMaker, aControl)
      else if aControl is TKnobsWirePanel
      then ShowPopupEditor( PopupMenuEditor, aControl)
      else if aControl is TKnobsModule
      then ShowPopupModule( PopupMenuModule, aControl)
      else if aControl is TKnobsSelector
      then ShowPopupSelector( PopupMenuSelector, aControl)
      else if aControl is TKnobsValuedControl
      then ShowPopupKnob( PopupMenuKnob, aControl)
      else if aControl is TKnobsXYControl
      then ShowPopupXYControl( PopupMenuXYControl, aControl)
      else if aControl is TKnobsGridControl
      then ShowPopupSimple( PopupMenuGridControl, aControl)
      else if aControl is TKnobsTextControl
      then ShowPopupTextControl( PopupMenuTextControl, aControl);
    end;


    procedure   TFormAbstraction.DoUnFocus( const aSender: TObject);
    begin
      // todo : implementation needed
    end;


    function    TFormAbstraction.DoLoadDataGraph( const aSender: TObject): string;
    begin
      // todo : implementation needed
    end;


    procedure   TFormAbstraction.DoSaveDataGraph( const aSender: TObject; const aValue: string);
    begin
      // todo : implementation needed
    end;


    function    TFormAbstraction.DoLoadGridControl( const aSender: TObject): string;
    begin
      // todo : implementation needed
    end;


    procedure   TFormAbstraction.DoSaveGridControl( const aSender: TObject; const aValue: string);
    begin
      // todo : implementation needed
    end;


    procedure   TFormAbstraction.DoActiveVariationChanged( const aSender: TObject; aVariation: Integer);
    begin
      // We have no worm panel, nothing to do till we have
    end;


    procedure   TFormAbstraction.DoLoadCaptions( const aSender: TObject; const aModule: TKnobsCustomModule; const aControl: TKnobsSelector; const aDependency: TKnobsValuedControl);
    begin
      // todo : seems to work without loading captions?
    end;


    procedure   TFormAbstraction.DoEditHistoryChanged( aSender: TObject);
    begin
      // todo: edit history not implemente yet
    end;


    procedure   TFormAbstraction.DoModulesRenamed( const aSender: TObject; const anOldNames, aNewNames: Knobs2013.TStringArray);
    begin
      // not used in main editor either
    end;


    procedure   TFormAbstraction.DoModulesRenamedFriendly( const aSender: TObject; const anOldNames, aNewNames: Knobs2013.TStringArray);
    begin
      // not used in main editor either
    end;


    procedure   TFormAbstraction.DoAudioChanged( const aSender: TObject; const anOldFileName, aNewFileName: string);
    var
      aCache : TAudioCache;
    begin
      aCache := GetAudioCache;

      if   Assigned( aCache)
      then begin
        if   anOldFileName <> aNewFileName
        then begin
          aCache.AddWaveInfo ( aNewFileName );
          aCache.FreeWaveInfo( anOldFileName);
        end;
      end;
    end;


    procedure   TFormAbstraction.DoBeginStateChange( const aSender: TObject);
    begin
      // todo : work ...
//      if   FStateChangeCount = 0
//      then begin
//        if   Assigned( FSynthesizer)
//        then begin
//          FAudioRefList := CreateAudioRefList;
//          FSynthesizer.CollectAudioRefs( FAudioRefList);
//
//          if   FAudioRefList.Count > 0
//          then IncAudioRefCountsFor( FAudioRefList)
//          else FreeAndNil( FAudioRefList);
//        end
//        else FreeAndNil( FAudioRefList);
//      end;
//
//      Inc( FStateChangeCount);
    end;


    procedure   TFormAbstraction.DoEndStateChange( const aSender: TObject);
    begin
      // todo : work ...
//      Dec( FStateChangeCount);
//
//      if   FStateChangeCount = 0
//      then begin
//        if   Assigned( FAudioRefList)
//        then begin
//          if   FAudioRefList.Count > 0
//          then DecAudioRefCountsFor( FAudioRefList);
//
//          FreeAndNil( FAudioRefList);
//        end;
//      end;
    end;


    procedure   TFormAbstraction.SelectNone;
    begin
      PatchEditor.UnselectAllModules;
    end;


    procedure   TFormAbstraction.AddModule( aModuleType: TKnobsModuleType);
    begin
      if Assigned( FormStore)
      then begin
        PatchEditor.BeginStateChange( True);

        try
          SelectNone;
          FormStore.CreateModule( PatchEditor, aModuleType, DO_DRAG, StandardColors).FixupInsertion;
        finally
          PatchEditor.EndStateChange( True, True);
        end;
      end;
    end;


    procedure   TFormAbstraction.ChangeModuleType( var aTarget: TKnobsCustomModule; aWantedType: TKnobsModuleType);
    // User requested a module type change trough the module popup menu.
    //
    //   get connections for aTarget
    //   remove MIDI stuff for aTarget
    //   remove aTarget
    //   Insert aWantedType with AddModule
    //   connect new module with the obtained connections
    //   copy over same named parameters
    //
    // NOTE : aTarget is freed and set to Nil in the process, do not use it after
    var
      aWire         : TKnobsWire;
      aConnections  : TStringList;
      aModule       : TKnobsCustomModule;
      anOldName     : string;
      aNewName      : string;
      anOldTypeName : string;
      aNewTypeName  : string;
      anOldType     : string;
      aNewType      : string;
      i             : Integer;
      aConnection   : string;
      aParams       : string;
    begin
      if   Assigned( aTarget)
      then begin
        PatchEditor.BeginStateChange( True);
        aTarget.SelectUnique;
        aConnections := TStringList.Create;

        try
          anOldType     := IntToStr( aTarget.ModuleType);
          aNewType      := IntToStr( aWantedType);
          anOldName     := aTarget.Name;
          anOldTypeName := LookupModuleTypeName( aTarget.ModuleType);

          for i := 0 to PatchEditor.WireCount - 1                                                          // Obtain wires
          do begin
            aWire := PatchEditor.Wire[ i];

            if   aWire.Involves( aTarget)
            then aConnections.Add( aWire.AsCompactString);
          end;

          aParams := FPatchWriter.WriteString( PatchEditor, '', wmSelected);                               // Obtain parameter settings
          aModule := FormStore.ChangeModule( PatchEditor, aTarget, aWantedType, DoOnValuedControlRemoved); // NILs aTarget and unassigns MIDI cintrol, do not use it after here
          aModule.FixupInsertion;
          aNewName     := aModule.Name;                                                                    // The new module name
          aNewTypeName := LookupModuleTypeName( aWantedType);                                              // The new module type name
          aModule.SelectUnique;
          FPatchReader.ReadParams( aParams, PatchEditor, NO_STRICT);                                       // Fix parameter settings

          for i := 0 to aConnections.Count - 1                                                             // Restore wires
          do begin
            aConnection := ReplaceText( aConnections[ i], Format( '%s.%s', [ anOldName, anOldTypeName], AppLocale) , Format( '%s.%s', [ aNewName, aNewTypeName], AppLocale));
            PatchEditor.AddWire( aConnection);
          end;
        finally
          aConnections.DisposeOf;
          PatchEditor.EndStateChange( True, False);
        end;
      end;
    end;


//  public

    procedure   TFormAbstraction.MakeSpedUpSlow;
    begin
      PatchEditor.MakeSpedUpSlow;
    end;


    procedure   TFormAbstraction.FixAllWires;
    begin
      PatchEditor.FixAllWires;
    end;


    procedure   TFormAbstraction.SortGraph;
    begin
      PatchEditor.SortGraph;
    end;


    procedure   TFormAbstraction.BuildModuleIndex;
    begin
      PatchEditor.BuildModuleIndex;
    end;


    procedure   TFormAbstraction.FixAllSynthParams( IsAutomation: Boolean);
    begin
      PatchEditor.FixAllSynthParams( IsAutomation);
    end;


    function    TFormAbstraction.FindModule( const aName: string): TKnobsCustomModule;
    begin
      Result := PatchEditor.FindModule( aName);
    end;


    function    TFormAbstraction.AbstractionToStr: string;
    begin
      Result := FPatchWriter.WriteString( PatchEditor);
    end;


    procedure   TFormAbstraction.LoadPatchFromString( const aPatch: string);
    begin
      FPatchReader.ReadString(
        aPatch         ,
        PatchEditor    ,
        Point( 0, 0)   ,
        DO_STRICT      ,
        NO_DRAG        ,
        DO_COPY_MIDI   ,
        DO_COPY_TUNING ,
        DO_COPY_RATE   ,
        rmReplace      ,
        ''
      )
    end;


    procedure   TFormAbstraction.BuildIoMaps( var aMaps: TKnobsIoArray);
    var
      InputNames  : TStringList;
      OutputNames : TStringList;

      function IsInput( aType: Integer): Boolean;
      begin
        Result := aType in [ MTYPE_AUDIO_IN, MTYPE_CONTROL_IN, MTYPE_LOGIC_IN, MTYPE_CONTROLLOGIC_IN];
      end;

    const
      CLEARANCE = 3; // connector to module border spacing
    var
      aForm        : TFormWren;
      aSynthModule : TModAbstraction;
      aName        : string;
      i            : Integer;
      c            : TKnobsConnector;
      s            : TSignalType;
      l            : TKnobsTextLabel;
      in_count     : Integer;
      out_count    : Integer;
      top          : Integer;
      top_input    : Integer;
      top_output   : Integer;
    begin
      aForm        := AsFormWren   ( FFormWren   );
      aSynthModule := AsSynthModule( FSynthModule);

      if   Assigned( FPatchModule)
      and  Assigned( aForm       )
      and  Assigned( aSynthModule)
      then begin

        // Clear what we have already

//        for i := FPatchModule.ControlCount - 1 downto 0
//        do begin
//          if   FPatchModule.Controls[ i] is TKnobsConnector
//          then FPatchModule.Controls[ i].DisposeOf
//          else if FPatchModule.Controls[ i] is TKnobsTextLabel
//          then FPatchModule.Controls[ i].DisposeOf;
//        end;


        InputNames  := TStringList.Create;
        OutputNames := TStringList.Create;

        try
          // Rebuild things

          PatchEditor.BuildIoMaps( aMaps);
          TArray.Sort< TKnobsIoRecord>(
            aMaps,
            TDelegatedComparer< TKnobsIoRecord>.Construct(
              function( const aLeft, aRight: TKnobsIoRecord): Integer
              begin
                Result := TComparer< Integer>.Default.Compare( aLeft.ConnectorType, aRight.ConnectorType);

                if   Result = 0
                then Result := TComparer< string>.Default.Compare( aLeft.Name, aRight.Name);
              end
            )
          );

          in_count   := 0;
          out_count  := 0;
          top_input  := 2 * MOD_Y_UNIT; // Leave some room at the top for the fixed module elements (edit button, title, etc.)
          top_output := 2 * MOD_Y_UNIT;
          top        := Max( top_input, top_output);

          for i := 0 to Length( aMaps) - 1
          do begin
            c := nil;
            l := nil;
            s := stControl;

            case aMaps[ i].ConnectorType of
              MTYPE_AUDIO_IN         : begin c := TKnobsInput .Create( nil); s := stAudio;      Inc( in_count ); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_AUDIO_OUT        : begin c := TKnobsOutput.Create( nil); s := stAudio;      Inc( out_count); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_CONTROL_IN       : begin c := TKnobsInput .Create( nil); s := stControl;    Inc( in_count ); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_CONTROL_OUT      : begin c := TKnobsOutput.Create( nil); s := stControl;    Inc( out_count); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_LOGIC_IN         : begin c := TKnobsInput .Create( nil); s := stLogic;      Inc( in_count ); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_LOGIC_OUT        : begin c := TKnobsOutput.Create( nil); s := stLogic;      Inc( out_count); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_CONTROLLOGIC_IN  : begin c := TKnobsInput .Create( nil); s := stContrLogic; Inc( in_count ); l := TKnobsTextLabel.Create( FPatchModule) end;
              MTYPE_CONTROLLOGIC_OUT : begin c := TKnobsOutput.Create( nil); s := stContrLogic; Inc( out_count); l := TKnobsTextLabel.Create( FPatchModule) end;
            end;

            if   Assigned( c)
            and  Assigned( l)
            then begin
              l.Parent          := FPatchModule;
              c.Parent          := FPatchModule;
              c.SignalType      := s;
              c.UseTypedBorders := Editor.UseTypedConnectorBorders;
              c.UseBorders      := Editor.UseConnectorBorders;

              if( IsInput( aMaps[ i].ConnectorType))
              then begin
                aName := format( 'abstraction_in%d', [ in_count], AppLocale);

                if   FPatchModule.CheckNameChange( c, c.Name, aName)
                then begin
                  c.Name := aName;
                  InputNames.Add( Format( 'in%d', [ in_count], AppLocale));
                end;

                c.Left    := CLEARANCE;
                c.Top     := top_input;
                l.Caption := Format( '%d - %s', [ in_count, aMaps[ i].Name], AppLocale);
                l.Left    := c.Left + c.Width + 2;
                l.Top     := c.Top - 1;
                Inc( top_input, MOD_Y_UNIT);
              end
              else begin
                aName     := format( 'abstraction_out%d', [ in_count], AppLocale);

                if   FPatchModule.CheckNameChange( c, c.Name, aName)
                then begin
                  c.Name := aName;
                  OutputNames.Add( Format( 'out%d', [ out_count], AppLocale));
                end;

                c.Left    := FPatchModule.Width - CLEARANCE - c.Width;
                c.Top     := top_output;
                l.Caption := Format( '%d - %s', [ out_count, aMaps[ i].Name], AppLocale);
                l.Left    := FPatchModule.Width div 2;
                l.Top     := c.Top - 1;
                Inc( top_output, MOD_Y_UNIT);
              end;

              top := Max( top_input, top_output);
            end;
          end;

          aSynthModule.InputNames  := InputNames;
          aSynthModule.OutputNames := OutputNames;
        finally
          FreeAndNil( InputNames );
          FreeAndNil( OutputNames);
        end;

        FPatchModule.Height := top; // todo: 'snap' this to a multiple of MOD_Y_UNIT (which is sort of auto-magic now)
      end;
    end;


// Delphi area

{$R *.dfm}

procedure TFormAbstraction.UpDownPolyphonyClick(Sender: TObject; Button: TUDBtnType);
begin
  if   Button = btNext
  then Polyphony := Polyphony + 1
  else Polyphony := Polyphony - 1;
end;

procedure TFormAbstraction.KnobsSelector1Click(Sender: TObject);
begin
  SavePatchPreset;
end;

procedure TFormAbstraction.abstraction_editClick(Sender: TObject);
begin
  LoadPatchPreset;
end;

procedure TFormAbstraction.MenuItemModuleColorClick(Sender: TObject);
begin
  ChangeModuleColor( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleCopyClick(Sender: TObject);
begin
  ModuleCopy( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleCutClick(Sender: TObject);
begin
  ModuleCut( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleDefaultcolorClick(Sender: TObject);
begin
  ModuleColorDefault( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleDeleteClick(Sender: TObject);
begin
  ModuleDelete( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleExludeFromRandomizationClick(Sender: TObject);
begin
  ModuleRandomizationToggle( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleHelpClick(Sender: TObject);
begin
  ShowModuleHelp( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleLoadFromPresetClick(Sender: TObject);
begin
  if   FPopupSender is TKnobsCustomModule
  then LoadPreset( FPopupSender as TKnobsCustomModule);
end;

procedure TFormAbstraction.MenuItemModulePasteParamsStrictClick(Sender: TObject);
begin
  ModulePasteParams( FPopupSender as TKnobsModule, True);
end;

procedure TFormAbstraction.MenuItemModuleRandomizeClick(Sender: TObject);
begin
  ModuleRandomize( FPopupSender as TKnobsModule);
end;

procedure TFormAbstraction.MenuItemModuleSaveAsPresetClick(Sender: TObject);
begin
  if   FPopupSender is TKnobsCustomModule
  then SavePreset( FPopupSender as TKnobsCustomModule);
end;

procedure TFormAbstraction.MenuItemPasteParamsNonStrictClick(Sender: TObject);
begin
  ModulePasteParams( FPopupSender as TKnobsModule, False);
end;

procedure TFormAbstraction.Customwirecolor1Click(Sender: TObject);
begin
  SetWireColorCustom( FPopupSender as TKnobsConnector);
end;

procedure TFormAbstraction.Makegreen1Click(Sender: TObject);
begin
  SetWireColor( FPopupSender as TKnobsConnector, clLime);
end;

procedure TFormAbstraction.Yellowwires1Click(Sender: TObject);
begin
  SetWireColor( FPopupSender as TKnobsConnector, clYellow);
end;

procedure TFormAbstraction.Redwires1Click(Sender: TObject);
begin
  SetWireColor( FPopupSender as TKnobsConnector, clRed);
end;

procedure TFormAbstraction.Bluewires1Click(Sender: TObject);
begin
  SetWireColor( FPopupSender as TKnobsConnector, clBlue);
end;

procedure TFormAbstraction.BreakWireClick(Sender: TObject);
begin
  ConnectorBreak( FPopupSender as TKnobsConnector);
end;

procedure TFormAbstraction.Defaultwirecolor1Click(Sender: TObject);
begin
  SetWireColorDefault( FPopupSender as TKnobsConnector);
end;

procedure TFormAbstraction.DeleteConnectorClick(Sender: TObject);
begin
  ConnectorDelete( FPopupSender as TKnobsConnector);
end;

procedure TFormAbstraction.DisconnectConnectorClick(Sender: TObject);
begin
  ConnectorDisconnect( FPopupSender as TKnobsConnector);
end;

procedure TFormAbstraction.FormCreate(Sender: TObject);
begin
  FPatchReader                         := TPatchReader.Create;
  FPatchWriter                         := TPatchWriter.Create;
  PatchEditor.PatchReader              := FPatchReader;
  PatchEditor.PatchWriter              := FPatchWriter;
  PatchEditor.OnHistoryChange          := DoHistoryChanged;
  PatchEditor.OnValueChanged           := DoValueChanged;
  PatchEditor.OnTextChanged            := DoTextChanged;
  PatchEditor.OnRecompile              := DoRecompile;
  PatchEditor.OnShowPoup               := DoShowPopupMenu;
  PatchEditor.OnUnFocus                := DoUnFocus;
  PatchEditor.OnLoadDataGraph          := DoLoadDataGraph;
  PatchEditor.OnSaveDataGraph          := DoSaveDataGraph;
  PatchEditor.OnLoadGridControl        := DoLoadGridControl;
  PatchEditor.OnSaveGridControl        := DoSaveGridControl;
  PatchEditor.OnActiveVariationChanged := DoActiveVariationChanged;
  PatchEditor.OnLoadCaptions           := DoLoadCaptions;
  PatchEditor.OnEditHistoryChanged     := DoEditHistoryChanged;
  PatchEditor.OnModulesRenamed         := DoModulesRenamed;
  PatchEditor.OnModulesRenamedFriendly := DoModulesRenamedFriendly;
  PatchEditor.OnAudioChanged           := DoAudioChanged;
  PatchEditor.OnBeginStateChange       := DoBeginStateChange;
  PatchEditor.OnEndStateChange         := DoEndStateChange;
end;

procedure TFormAbstraction.FormDestroy(Sender: TObject);
begin
  AsSynthModule( FSynthModule).EditorDestroyed;
end;

procedure TFormAbstraction.ModuleSelectorModuleButtonClick(const aSender: TObject; aModuleclass: Integer);
begin
  AddModule( aModuleClass);
end;


end.
